import threading

class FlushTimer:
    def __init__(self, flush_fn, interval: int):
        self.flush_fn = flush_fn
        self.interval = interval
        self._timer: threading.Timer | None = None
        self._lock = threading.Lock()

    def schedule(self):
        with self._lock:
            if self._timer and self._timer.is_alive():
                return
            self._timer = threading.Timer(self.interval, self.flush_fn)
            self._timer.start()



import sys
import traceback
import asyncio

def intercept_exceptions(logger):
    original_hook = sys.excepthook

    def custom_hook(exc_type, exc_value, exc_traceback):
        ts = logger.options.dynamic_labels.get("timestamp", lambda: "")()
        error_message = ''.join(traceback.format_exception(exc_type, exc_value, exc_traceback))
        logger.log_buffer.append((ts, f"[EXCEPTION] {error_message}"))
        try:
            loop = asyncio.get_event_loop()
            if loop.is_running():
                loop.create_task(logger.flush_logs())
            else:
                asyncio.run(logger.flush_logs())
        except Exception:
            pass  

        original_hook(exc_type, exc_value, exc_traceback)

    sys.excepthook = custom_hook



import aiohttp

from .utils import get_labels


async def post_to_loki(log_buffer, options):
    payload = {
        "streams": [
            {
                "stream": get_labels(options),
                "values": log_buffer,
            }
        ]
    }
    headers = {
        "X-Scope-OrgID": options.tenant_id,
    }
    if options.auth_token:
        headers["Authorization"] = f"Bearer {options.auth_token}"

    try:
        async with aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=3)) as session:
            async with session.post(options.url, json=payload, headers=headers) as response:
                if response.status != 204:
                    await response.text()
    except Exception:
        pass



import logging


class LokiHandler(logging.Handler):
    def __init__(self, logger):
        super().__init__()
        self.logger = logger

    def emit(self, record):
        try:
            ts = self.logger.options.dynamic_labels.get("timestamp", lambda: "")()
            msg = self.format(record)
            self.logger.log_buffer.append((ts, f"[{record.levelname}] {msg}"))
            self.logger.flush_manager.check_and_flush()
        except Exception:
            pass

def intercept_logging(logger):
    handler = LokiHandler(logger)
    handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)
    root_logger.addHandler(handler)



from .console import intercept_print
from .exceptions import intercept_exceptions
from .logging import intercept_logging

__all__ = [
    "intercept_print",
    "intercept_exceptions",
    "intercept_logging",
]



import asyncio

from .aiohttp_session import post_to_loki


class AsyncFlushManager:
    def __init__(self, logger):
        self.logger = logger
        self._flush_task: asyncio.Task | None = None

    def check_and_flush(self):
        if len(self.logger.log_buffer) >= self.logger.options.batch_size:
            asyncio.create_task(self.logger.flush_logs())
        else:
            self._schedule_flush()

    def _schedule_flush(self):
        if self._flush_task and not self._flush_task.done():
            return
        self._flush_task = asyncio.create_task(self._delayed_flush())

    async def _delayed_flush(self):
        await asyncio.sleep(self.logger.options.flush_interval)
        await self.logger.flush_logs()

    async def send_logs(self, buffer_copy):
        await post_to_loki(buffer_copy, self.logger.options)



import asyncio
import atexit

from .config import LokiLoggerOptions
from .flush_manager import AsyncFlushManager
from .interceptors import (intercept_exceptions, intercept_logging,
                           intercept_print)
from .utils import now_ns, safe_json


class LokiLogger:
    def __init__(self, options: LokiLoggerOptions):
        self.options = options
        self.log_buffer: list[tuple[str, str]] = []
        self.flush_manager = AsyncFlushManager(self)
        intercept_print(self)
        intercept_logging(self)
        intercept_exceptions(self)
        atexit.register(lambda: asyncio.run(self.flush_logs()))

    def track_event(self, event_name: str, properties: dict | None = None) -> None:
        ts = now_ns()
        message = f"[EVENT] {event_name} {safe_json(properties)}" if properties else f"[EVENT] {event_name}"
        self.log_buffer.append((ts, message))
        self.flush_manager.check_and_flush()

    async def flush_logs(self) -> None:
        if not self.log_buffer:
            return
        buffer_copy = self.log_buffer.copy()
        self.log_buffer.clear()
        await self.flush_manager.send_logs(buffer_copy)



import builtins


def intercept_print(logger):
    original_print = builtins.print

    def custom_print(*args, **kwargs):
        ts = logger.options.dynamic_labels.get("timestamp", lambda: "")()
        message = " ".join(str(a) for a in args)
        logger.log_buffer.append((ts, f"[PRINT] {message}"))
        logger.flush_manager.check_and_flush()
        original_print(*args, **kwargs)

    builtins.print = custom_print



from typing import Callable, Dict, Optional, Any

class LokiLoggerOptions:
    def __init__(
        self,
        url: str,
        tenant_id: str,
        app_name: str,
        auth_token: Optional[str] = None,
        batch_size: int = 10,
        flush_interval: int = 2,
        labels: Optional[Dict[str, str]] = None,
        dynamic_labels: Optional[Dict[str, Callable[[], Any]]] = None,
    ):
        self.url = url
        self.tenant_id = tenant_id
        self.app_name = app_name
        self.auth_token = auth_token
        self.batch_size = batch_size
        self.flush_interval = flush_interval
        self.labels = labels or {}
        self.dynamic_labels = dynamic_labels or {}



import json
import time


def now_ns() -> str:
    return str(int(time.time() * 1_000_000_000))

def safe_json(data) -> str:
    try:
        return json.dumps(data)
    except Exception:
        return "{}"

def get_labels(options):
    try:
        dynamic_labels = {
            k: str(v()) for k, v in options.dynamic_labels.items()
        }
    except Exception:
        dynamic_labels = {}
    return {
        "app": options.app_name,
        **options.labels,
        **dynamic_labels,
    }
